"""A few default JIRA fields with their definitions and formats."""
import operator

from jirajumper.fields.field import JiraField
from jirajumper.fields.repository import JiraFieldsRepository
from jirajumper.models import FieldByName

get_name = operator.attrgetter('name')


VERSION = JiraField(
    jira_name='fixVersions',
    human_name='version',
    description='Software product version the issue is attached to.',

    to_jira=lambda version_name: [{'name': version_name}],
    from_jira=lambda jira_value: (jira_value or None) and jira_value[0].name,
)

SUMMARY = JiraField(
    jira_name='summary',
    human_name='summary',
    description='Issue summary (issue name, issue title).',
)


EPIC_LINK = JiraField(
    jira_name=FieldByName('Epic Link'),
    human_name='epic',
    description='Epic which the issue belongs to.',
)


STATUS = JiraField(
    jira_name='status',
    human_name='status',
    description='Issue status.',

    # It is impossible to set issue status directly in update() operation.
    to_jira=NotImplemented,
    from_jira=get_name,
)


TYPE = JiraField(
    jira_name='issuetype',
    human_name='type',
    description='Issue type.',

    to_jira=lambda type_name: {'name': type_name},
    from_jira=get_name,
)

PROJECT = JiraField(
    jira_name='project',
    human_name='project',
    description='JIRA Project.',

    # It is impossible to easily migrate across projects.
    to_jira=NotImplemented,
    from_jira=get_name,
)

DESCRIPTION = JiraField(
    jira_name='description',
    human_name='description',
    description='JIRA task description body.',
)


# TODO This should be configurable in configuration files or somehow else.
FIELDS = JiraFieldsRepository([
    VERSION, SUMMARY, STATUS, TYPE, EPIC_LINK, PROJECT, DESCRIPTION,
])



import os
from typing import Optional

from jira import JIRA

from jirajumper.errors import MissingJiraCredentials


def env_server() -> Optional[str]:
    """Retrieve JIRA server address."""
    return os.getenv('JIRA_SERVER')


def env_username() -> Optional[str]:
    """Retrieve JIRA server username."""
    return os.getenv('JIRA_USERNAME')


def env_token() -> Optional[str]:
    """Retrieve JIRA server token."""
    return os.getenv('JIRA_TOKEN')


def jira() -> JIRA:
    """Construct an instance of Jira client and establish an HTTP connection."""
    server = env_server()
    username = env_username()
    token = env_token()

    if not all([server, username, token]):
        raise MissingJiraCredentials(
            server=server,
            username=username,
            token=token,
        )

    return JIRA(
        server=server,
        basic_auth=(
            username,
            token,
        ),
    )


def issue_url(server: str, key: str) -> str:
    """Generate URL for a given issue."""
    return f'{server}/browse/{key}'


def custom_fields(client: JIRA):
    fields = client.fields()

    for field in fields:
        print(field)



from dataclasses import dataclass
from typing import Dict

import rich
from documented import DocumentedError
from jira import JIRAError

from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.fields import JiraFieldsRepository


@dataclass
class JIRAUpdateFailed(DocumentedError):
    """
    Cannot update a JIRA issue :(

    Errors
    {self.formatted_errors}
    """

    errors: Dict[str, str]
    fields: JiraFieldsRepository

    @property
    def formatted_errors(self) -> str:
        """Format the error list received from JIRA."""
        error_by_field = [
            (self.fields.find_by_jira_name(jira_name), error_message)
            for jira_name, error_message in self.errors.items()
        ]

        return '\n'.join([
            f'  - {field.human_name}\n      {error_message}'
            for field, error_message in error_by_field
        ])


def update(
    context: JeevesJiraContext,
    **kwargs: str,
):
    """
    Update the selected JIRA issue.

    Use `jj jump` to select the issue to update.
    """
    fields_and_values = [
        (applicable_field, kwargs[applicable_field.human_name])
        for applicable_field in context.obj.fields
        if kwargs.get(applicable_field.human_name)
    ]

    rich.print('Updating:')
    for print_field, human_value in fields_and_values:
        rich.print(f'  - {print_field.human_name} ≔ {human_value}')

    issue_fields = dict([
        store_field.store(human_value)
        for store_field, human_value in fields_and_values
    ])

    try:
        context.obj.current_issue.update(issue_fields)
    except JIRAError as err:
        raise JIRAUpdateFailed(
            errors=err.response.json().get('errors', {}),
            fields=context.obj.fields,
        ) from err

    rich.print('Updated!')



import json
from typing import Optional

import backoff
import rich
from documented import DocumentedError
from jira import JIRA, JIRAError
from typer import Argument, echo

from jirajumper.cache.cache import JeevesJiraContext, JiraCache
from jirajumper.client import issue_url
from jirajumper.models import OutputFormat


def normalize_issue_specifier(
    client: JIRA,
    specifier: str,
    current_issue_key: Optional[str],
):
    """Normalize issue specifier."""
    if specifier.isnumeric() and current_issue_key:
        project_key, _current_issue_number = current_issue_key.split('-')
        return f'{project_key}-{specifier}'

    if specifier.lower() == 'next':
        current_issue = client.issue(current_issue_key)
        links = current_issue.fields.issuelinks

        if not links:
            raise ValueError(
                f'Issue {current_issue_key} does not have any issues it blocks.',
            )

        for link in links:
            try:
                outward_issue = link.outwardIssue
            except AttributeError:
                continue

            if (
                link.type.name == 'Blocks' and
                outward_issue.fields.status.statusCategory.name != 'Done'
            ):
                return outward_issue.key

        raise ValueError(
            f'Cannot find an issue that follows {current_issue_key} :(',
        )

    return specifier


class NoIssueSelected(DocumentedError):
    """
    No issue has been selected.

    To select an issue PROJ-123, please run:

        jj jump PROJ-123
    """


@backoff.on_exception(backoff.expo, JIRAError, max_time=5)
def jump(
    context: JeevesJiraContext,
    specifier: Optional[str] = Argument(None),    # noqa: WPS404
):
    """Select a Jira issue to work with."""
    client = context.obj.jira
    cache = context.obj.cache

    if specifier:
        specifier = normalize_issue_specifier(
            client=client,
            specifier=specifier,
            current_issue_key=cache.selected_issue_key,
        )

        issue = client.issue(specifier)
        cache.selected_issue_key = issue.key

        context.obj.store_cache(
            JiraCache(
                selected_issue_key=issue.key,
            ),
        )
    else:
        key = cache.selected_issue_key

        if not key:
            raise NoIssueSelected()

        issue = client.issue(cache.selected_issue_key)

    if context.obj.output_format == OutputFormat.PRETTY:
        rich.print(f'[bold]{issue.key}[/bold] {issue.fields.summary}')
        rich.print(issue_url(client.server_url, issue.key))

        for print_field in context.obj.fields:
            field_value = print_field.retrieve(
                issue=issue,
                field_key_by_name=context.obj.field_key_by_name,
            )
            rich.print(f'  - {print_field.human_name}: {field_value}')

    else:
        echo(
            json.dumps(
                {
                    json_field.human_name: json_field.retrieve(
                        issue=issue,
                        field_key_by_name=context.obj.field_key_by_name,
                    )
                    for json_field in context.obj.fields
                },
                indent=2,
            ),
        )

    return issue



import json
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional

from jira import JIRA, Issue
from pydantic import BaseModel, Field
from typer import Context

from jirajumper.fields import JiraFieldsRepository
from jirajumper.fields.field import FieldKeyByName
from jirajumper.models import OutputFormat


class IssueFieldSchema(BaseModel):
    """JIRA issue field schema."""

    field_type: str = Field(alias='type')
    system: Optional[str] = None


class JiraCache(BaseModel):
    """Cached JIRA configuration."""

    selected_issue_key: Optional[str] = None


@dataclass
class GlobalOptions:
    """Global jeeves-jira configuration options."""

    output_format: OutputFormat
    jira: JIRA
    cache_path: Path
    fields: JiraFieldsRepository

    @cached_property
    def cache(self) -> JiraCache:
        """Retrieve jirajumper cache from disk."""
        try:
            raw_cache = self.cache_path.read_text()
        except (FileNotFoundError, PermissionError):
            return JiraCache()

        try:
            return JiraCache.parse_raw(raw_cache)
        except ValueError:
            return JiraCache()

    def store_cache(self, cache: JiraCache):
        """Store cache contents on disk."""
        encoded_cache = cache.json(by_alias=True)

        try:
            self.cache_path.write_text(encoded_cache)
        except FileNotFoundError:
            self.cache_path.parent.mkdir(parents=True, exist_ok=True)
            self.cache_path.write_text(encoded_cache)

    @cached_property
    def current_issue(self) -> Issue:
        """Construct the currently selected JIRA issue object."""
        return self.jira.issue(self.cache.selected_issue_key)

    @cached_property
    def field_key_by_name(self) -> FieldKeyByName:
        """
        Map field names to keys.

        This is useful for `custom*` and the like: the field key can vary
        among installations but the canonical name helps us to find that key.
        """
        return {
            field['name']: field['key']
            for field in self.jira.fields()
        }


class JeevesJiraContext(Context):
    """Typer context with GlobalOptions instance as obj."""

    obj: GlobalOptions



from typing import Optional

import rich
from typer import Option

from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.models import CloneIssue


def clone(
    context: JeevesJiraContext,

    summary: str,
    epic_link: Optional[str] = Option(None, '--epic'),
    issue_type: Optional[str] = Option(None, '--type'),
    version: Optional[str] = None,
    project: Optional[str] = None,
    stay: bool = Option(
        False,
        help=(
            'Stay on current issue instead of switching '
            'to the newly created one'
        ),
    ),
    epic_name: Optional[str] = None,
    assignee: Optional[str] = None,
    status: Optional[str] = None,
):
    """Clone a JIRA issue."""
    fields = CloneIssue(
        summary=summary,
        issue_type=issue_type,
        version=version,
        project=project,
        epic_link=epic_link,
        epic_name=epic_name,
        assignee=assignee,
    )

    client = context.obj.jira
    cache = context.obj.cache

    parent_issue = client.issue(cache.selected_issue_key)

    if assignee:
        assignee_account_id = find_account_id_by_user_name(
            client=client,
            user_name=assignee,
        )

    elif parent_assignee := parent_issue.fields.assignee:
        assignee_account_id = parent_assignee.accountId

    else:
        assignee_account_id = None

    if parent_issue.fields.issuetype.name == 'Epic':
        if not epic_name:
            raise ValueError(
                'You are cloning an epic. Please specify --epic-name.',
            )

    project = project or parent_issue.fields.project.key

    description = parent_issue.fields.description

    status = parent_issue.fields.status.name
    epic = getattr(
        parent_issue.fields,
        epic_link_field,
    )
    parent = getattr(parent_issue.fields, 'parent', None)
    if parent and parent.fields.issuetype.name != 'Epic':
        parent_link = {
            'key': parent.key,
        }
    else:
        parent_link = None

    if not version and parent_issue.fields.fixVersions:
        version = parent_issue.fields.fixVersions[0].name

    if not version:
        raise ValueError(
            'No version specified in parent issue properties or '
            'in command line options.',
        )

    fields = {
        'parent': parent_link,
    }

    fields = dict(_prepare_fields(
        description=description,
        project=project,
        summary=summary,
        version=version,
        issue_type=issue_type or parent_issue.fields.issuetype.name,

        epic=epic,
        epic_name=epic_name,
        assignee_account_id=assignee_account_id,
    ))

    issue = client.create_issue(
        fields=fields,
    )

    if (
        issue.fields.status.name != status
        and parent_issue.fields.status.statusCategory.name != 'Done'
    ):
        transition_ids = [
            t['id'] for t in
            client.transitions(issue)
            if t['to']['name'] == status
        ]

        transition_id = first(transition_ids, None)

        if transition_id is None:
            rich.print(
                f'[orange]Warning:[/orange] Cannot transfer {issue} to status '
                f'[bold]{status}[/bold] because no direct transition found.',
            )

        else:
            client.transition_issue(issue, transition_id)

    if not stay:
        select(verbose=False, specifier=issue.key)

    else:
        rich.print(
            f'New issue: [bold]{issue.key}[/bold] {issue.fields.summary}')
        rich.print('  {}'.format(issue_url(client.server_url, issue.key)))
        rich.print(
            f'Selected issue: [bold]{parent_issue.key}[/bold] '
            f'{parent_issue.fields.summary}',
        )

    return issue




from pathlib import Path

import click
from typer import Context, Option, Typer
from typer.core import TyperCommand

from jirajumper.cache.cache import GlobalOptions
from jirajumper.client import jira
from jirajumper.commands.clone import clone
from jirajumper.commands.list_issues import list_issues
from jirajumper.commands.select import jump
from jirajumper.commands.update import update
from jirajumper.fields import FIELDS
from jirajumper.models import OutputFormat

app = Typer(
    help='Manage JIRA issues.',
    no_args_is_help=True,
)


@app.callback()
def global_options(
    context: Context,
    format: OutputFormat = Option(
        OutputFormat.PRETTY,
        help='Format to print the data in',
    ),
    cache_path: Path = Option(
        default=Path.home() / '.cache/jirajumper/jirajumper.json',
        envvar='JIRAJUMPER_CACHE_PATH',
        help='Path to the JSON file where jirajumper will store its cache.',
    ),
):
    """Configure global options valid for most of jeeves-jira commands."""
    context.obj = GlobalOptions(
        output_format=format,
        jira=jira(),
        fields=FIELDS,
        cache_path=cache_path,
    )


EDITABLE_FIELDS = frozenset({
    'epic',
    'type',
    'project',
    'version',
    'assignee',
    'description',
    'parent',
})


class UpdateCommand(TyperCommand):
    def __init__(self, **kwargs):
        custom_options = [
            click.Option(
                [f'--{field.human_name}'],
                help=field.description,
            )
            for field in FIELDS.writable()
        ]

        kwargs.update(
            params=custom_options,
        )

        super().__init__(**kwargs)


app.command()(jump)
app.command(name='list')(list_issues)
app.command()(clone)
app.command(
    cls=UpdateCommand,
    context_settings={
        'ignore_unknown_options': True,
    },
)(update)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jirajumper', 'jirajumper.cache', 'jirajumper.commands', 'jirajumper.fields']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.11.1,<2.0.0',
 'click<8.0.0',
 'documented>=0.1.1,<0.2.0',
 'jira>=3.0.1,<4.0.0',
 'more-itertools>=8.10.0,<9.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'rich>=10.12.0,<11.0.0',
 'stringcase>=1.2.0,<2.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['jj = jirajumper:app']}

setup_kwargs = {
    'name': 'jirajumper',
    'version': '0.1.0',
    'description': 'Yet another JIRA issue manager CLI with emphasis on task chains.',
    'long_description': '# jeeves-jira\n\n[![Build Status](https://github.com/jeeves-sh/jeeves-jira/workflows/test/badge.svg?branch=master&event=push)](https://github.com/jeeves-sh/jeeves-jira/actions?query=workflow%3Atest)\n[![codecov](https://codecov.io/gh/jeeves-sh/jeeves-jira/branch/master/graph/badge.svg)](https://codecov.io/gh/jeeves-sh/jeeves-jira)\n[![Python Version](https://img.shields.io/pypi/pyversions/jeeves-jira.svg)](https://pypi.org/project/jeeves-jira/)\n[![wemake-python-styleguide](https://img.shields.io/badge/style-wemake-000000.svg)](https://github.com/wemake-services/wemake-python-styleguide)\n\nYet another JIRA issue manager CLI with emphasis on task chains.\n\n\n## Features\n\n- Fully typed with annotations and checked with mypy, [PEP561 compatible](https://www.python.org/dev/peps/pep-0561/)\n- Add yours!\n\n\n## Installation\n\n```bash\npip install jirajumper\n```\n\n\n## License\n\n[MIT](https://github.com/jeeves-sh/jeeves-jira/blob/master/LICENSE)\n\n\n## Credits\n\nThis project was generated with [`wemake-python-package`](https://github.com/wemake-services/wemake-python-package). Current template version is: [86ec1f15472f7c9f634cdbaa0d33eddf166888b2](https://github.com/wemake-services/wemake-python-package/tree/86ec1f15472f7c9f634cdbaa0d33eddf166888b2). See what is [updated](https://github.com/wemake-services/wemake-python-package/compare/86ec1f15472f7c9f634cdbaa0d33eddf166888b2...master) since then.\n',
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/anatoly-scherbakov/jirajumper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Protocol, Tuple, TypeVar, Union

from jira import Issue

from jirajumper.models import (
    FieldByName,
    FieldKeyByName,
    HumanValue,
    JiraValue,
    NotImplementedType,
)

AnyType = TypeVar('AnyType')


class ToJira(Protocol):
    """Convert a human-readable value to JIRA field."""

    def __call__(self, human_value: HumanValue) -> JiraValue:
        """Convert a human-readable value to JIRA field."""
        raise NotImplementedError()


class FromJira(Protocol):
    """Convert a native JIRA field value to a human readable one."""

    def __call__(self, jira_value: JiraValue) -> HumanValue:
        """Convert a native JIRA field value to a human readable one."""
        raise NotImplementedError()


def identity(any_value: AnyType) -> AnyType:
    """Identity function."""
    return any_value


@dataclass(frozen=True)
class JiraField:
    """Transformations related to a particular JIRA field."""

    jira_name: Union[str, FieldByName]
    human_name: str
    description: str

    to_jira: Union[ToJira, NotImplementedType] = identity
    from_jira: FromJira = identity

    def resolve_jira_field_name(self, field_key_by_name: FieldKeyByName) -> str:
        """Resolve JIRA field name."""
        if isinstance(self.jira_name, str):
            return self.jira_name

        if isinstance(self.jira_name, FieldByName):
            return field_key_by_name[self.jira_name.readable_name]

        raise ValueError(f'`{self.jira_name}` is not a valid JIRA field name.')

    def retrieve(self, issue: Issue, field_key_by_name: FieldKeyByName):
        """Retrieve the native field value from given issue."""
        return self.from_jira(
            getattr(
                issue.fields,
                self.resolve_jira_field_name(
                    field_key_by_name=field_key_by_name,
                ),
            ),
        )

    def store(self, human_value: HumanValue) -> Tuple[str, JiraValue]:
        """Convert the readable value into JIRA native form."""
        return self.jira_name, self.to_jira(human_value)

    def is_writable(self):
        """Find out if this field is writable."""
        return self.to_jira is not NotImplemented



from enum import Enum
from typing import Any, Dict, NamedTuple, NewType, Optional, Union

from pydantic import BaseModel

FieldKeyByName = Dict[str, str]
JiraValue = NewType('JiraValue', Any)
HumanValue = NewType('HumanValue', Optional[Union[str, int, float]])
NotImplementedType = type(NotImplemented)


class OutputFormat(str, Enum):  # noqa: WPS600
    """Output format."""

    PRETTY = 'pretty'
    JSON = 'json'


class Fields(BaseModel):
    """Supported base fields."""

    summary: Optional[str] = None
    issue_type: Optional[str] = None
    version: Optional[str] = None
    project: Optional[str] = None
    epic_name: Optional[str] = None
    epic_link: Optional[str] = None
    assignee: Optional[str] = None


class CloneIssue(Fields):
    """Fields to clone an issue."""

    summary: str


class RetrieveIssue(CloneIssue):
    """Retrieved issue."""

    summary: str


class FieldByName(NamedTuple):
    """Find a field by its user-facing JIRA name."""

    readable_name: str



from typing import List, Optional

from more_itertools import first

from jirajumper.fields.field import JiraField


class JiraFieldsRepository(List[JiraField]):
    """List of supported JIRA issue fields."""

    def find_by_jira_name(self, jira_name: str) -> Optional[JiraField]:
        """Find a field by JIRA name."""
        return first(
            filter(
                lambda field: field.jira_name == jira_name,
                self,
            ), None,
        )

    def writable(self) -> 'JiraFieldsRepository':
        """Show only fields that are writable."""
        return JiraFieldsRepository(filter(
            JiraField.is_writable,
            self,
        ))
